import numpy as np
import pandas as pd
import sys, os, pickle
import argparse, textwrap
from .preprocess import data_preparation
from .common import  build_training_model, transfer_test
from .SHAP_importance import lauch_SHAP, sum_SHAP_values_of_genes, intersect_of_top_genes

def main():
    parser = argparse.ArgumentParser(description = 'Pipeline for building clinical outcome prediction models on training dataset and transfer learning on validation datasets.',
            usage = 'use "%(prog)s --help" for more information',
            formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--train_path', type = str,
            help = "Path to your training data, in .csv format; includes sample names as first column and labels as last column")
    parser.add_argument('--valid_path', type = str,
            help = "Path to your transfer validation data, in .csv format; includes sample names as first column and labels as last column")
    parser.add_argument('-m','--model_type', type = str,
            help = '''Machine learning models to use:
            lgb: LightGBM;
            xgb: XGBoost;
            rf: Random Forest;
            gpr: Gaussian Process Regression;
            lr: Linear Regression;
            default: lgb''',
            default = 'lgb')
    parser.add_argument('--no_quantile',
            action = "store_true",
            help = "If specified, do not use quantile normalization.")
    parser.add_argument('--shap',
            action = 'store_true',
            help = '''
            Conduct SHAP analysis on the validation set.
            Only for use with LightGBM, XGBoost, and Random Forest.
            ''')
    parser.add_argument('-n', '--top_genes', type = int,
            help = '''
            If --shap is specified, indicate number of top genes from both training and validation sets that will be compared in post-SHAP analysis.
            Default is 20.
            ''',
            default = 20)


    args = parser.parse_args()

    assert args.model_type in ['lgb','rf','xgb','gpr','lr'], "Model " + args.model_type + " not supported in Ciclops! Use ciclops --help for more information."

    if args.no_quantile:
        print("Do not use normalization ...")

    if args.shap:
        if args.model_type in ['lgb','xgb','rf']:
            print('Perform SHAP analysis after model training.')
            
    opts = vars(parser.parse_args())

    run(**opts)


def run(train_path, valid_path, model_type, no_quantile, shap, top_genes):
    # load training dataset
    df_train = pd.read_csv(train_path) # training dataset for building the model
    # load transfer validation dataset
    df_val = pd.read_csv(valid_path) # validation dataset for transfer testing

    # preprocess both training and validation dataset
    df_training, df_validation = data_preparation(df_train, df_val, no_quantile)
    generate_results(df_training, df_validation, model_type, '')

    if shap:
        if model_type in ['lgb','xgb','rf']:
            print("Launching SHAP analysis ...")
            lauch_SHAP()
            sum_SHAP_values_of_genes()
            intersect_of_top_genes(top_genes)
        else:
            print('--shap was specified, but since the model type is ' + model_type + ', no SHAP analysis will be performed.' )


def generate_results(df_training, df_validation, model_type, path_affix):
    """
    Params
    ------
    df_training: a Pandas DataFrame
    df_validation: a Pandas DataFrame
    model_type: str
    path_affix: str
    """
    #Part 1: 10-fold cross validation on training data
    eva_cv, eva_cv_conf = build_training_model(df_training, model_type)
    # Part 2: transfer testing on validation data
    eva_tv, eva_tv_conf = transfer_test(df_validation)

    # save evaluation results
    path = './'+path_affix+'performance'
    os.makedirs(path, exist_ok = True)
    eva_cv.to_csv(path+'/training_cv_results.csv', index = False)
    eva_cv_conf.to_csv(path+'/training_cv_confidence.csv', index = False)
    eva_tv.to_csv(path+'/validation_tv_results.csv', index = False)
    eva_tv_conf.to_csv(path+'/validation_tv_confidence.csv', index = False)

if __name__ == "__main__":
    main()



import numpy as np
import sklearn.ensemble
# light GBM modules
import lightgbm as lgb
# XGboost modules
from xgboost import XGBRegressor
# GPR modules
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
# linear regression modules
from sklearn.linear_model import LinearRegression
# random forest modules
import sklearn.ensemble

""" All types of model tested in this study
"""

def train_lighgbm_model(TRAIN):
    """ Train LightGBM Regression model
    Params:
    -------
    TRAIN: Pandas dataframe
        training set; last column is label
    Yields:
    -------
    lgb: LightGBM model
    """
    train_dat=TRAIN.iloc[:,:-1]
    train_gs=TRAIN.iloc[:,-1]
    lgb_train = lgb.Dataset(np.asarray(train_dat), np.asarray(train_gs))
    params = {
            'boosting_type': 'gbdt',
            'objective': 'regression',
            'num_leaves': 5,
            'learning_rate': 0.05,
            'verbose': 0,
            'n_estimators': 800,
            'reg_alpha': 2.0,
            #'first_metric_only': True
            }
    gbm = lgb.train(params,
            train_set = lgb_train,
            #valid_sets = lgb_test,
            #early_stopping_rounds = 5000,
            num_boost_round=1000
            )
    return gbm

def train_xgboost_model(TRAIN):
    """ Train XGboost Regression model
    Params:
    -------
    TRAIN: Pandas dataframe
        training set; last column is label
    Yields:
    -------
    xgp: XGboost Regression model
    """
    train_dat=TRAIN.iloc[:,:-1]
    train_gs=TRAIN.iloc[:,-1]
    xgb = XGBRegressor()
    xgb.fit(train_dat, train_gs)
    return xgb

def train_rf_model(TRAIN):
    """ Train Random Forest Regression model
    Params:
    -------
    TRAIN: Pandas dataframe
        training set; last column is label
    Yields:
    -------
    clf: Random Forest Regression model
    """
    train_dat=TRAIN.iloc[:,:-1]
    train_gs=TRAIN.iloc[:,-1]
    clf = sklearn.ensemble.RandomForestRegressor(n_estimators=100, max_depth=2, random_state=0)
    clf.fit(train_dat, train_gs)
    return clf


def train_gpr_model(TRAIN):
    """ Train Gaussian Process Regression model
    Params:
    -------
    TRAIN: Pandas dataframe
        training set; last column is label
    
    Yields:
    -------
    gpr: Gaussian Process Regression model
    """
    train_dat=TRAIN.iloc[:,:-1]
    train_gs=TRAIN.iloc[:,-1]
    kernel = DotProduct() + WhiteKernel()
    gpr = GaussianProcessRegressor(kernel=kernel, random_state=0).fit(train_dat, train_gs)
    return gpr

def train_lr_model(TRAIN):
    """ Train Linear Regression model
    Params:
    -------
    TRAIN: Pandas dataframe
        training set; last column is label
    
    Yields:
    -------
    lr: linear regression model
    """
    train_dat=TRAIN.iloc[:,:-1]
    train_gs=TRAIN.iloc[:,-1]
    lr =  LinearRegression().fit(train_dat, train_gs)
    return lr



import numpy as np
import pandas as pd
from sklearn.model_selection import KFold
import os, pickle
from scipy.stats import pearsonr, spearmanr
from .models import *
from .statistics import *

def build_training_model(df_training, model_type):
    """ Build clinical outcome prediction model and 10-fold cross validation from training dataset.
    Model parameters for each fold are saved as fold_*_model.sav

    Parameters:
    -----------
    df_training: a Pandas dataframe
        the processed training dataset
    model_type: str
        type of model to train
        'lgb': Light GBM model
        'xgb': XGboost model
        'rf': Random Forest
        'lr': Linear regression
        'gpr' gaussian process regression

    Yields:
    -------
    eva_df: a Pandas dataframe
        evaluations of training models during 10-fold cross validation
    """
    os.makedirs('./training/', exist_ok = True) # where to save the processed training datasets during 10-fold cross validation
    os.makedirs('./params/', exist_ok = True) # where to save the model params
    kf = KFold(n_splits=10, shuffle = True, random_state = 0)
    eva_df = {'fold':[], 'AUROC':[], 'AUPRC':[], 'Pearsonr':[], 'Spearmanr':[]}
    eva_conf_df = {'Pearsonr mean[95CI]':[],'Spearmanr mean[95CI]':[],'AUROC mean[95CI]':[], 'AUPRC mean[95CI]':[]}

    pred_all = []
    gs_all = []

    for i, (train_idx,test_idx) in enumerate(kf.split(df_training)):
        print('Start preparing fold', i, '...')
        path = './training/fold_'+str(i)
        os.makedirs(path, exist_ok = True)
        TRAIN = df_training.iloc[train_idx]
        TEST = df_training.iloc[test_idx]
        TRAIN.to_csv(path+"/Train.csv", index = False)
        TEST.to_csv(path+"/Test.csv", index = False)
        TRAIN_data = TRAIN.iloc[:,1:] # assuming first column is the sample name/row numbers
        TEST_data = TEST.iloc[:,1:]

        # train model on training dataset
        if model_type == 'lgb':
            predictor = train_lighgbm_model(TRAIN_data)
        elif model_type == 'xgb':
            predictor = train_xgboost_model(TRAIN_data)
        elif model_type == 'rf':
            predictor = train_rf_model(TRAIN_data)
        elif model_type == 'lr':
            predictor = train_lr_model(TRAIN_data)
        elif model_type == 'gpr':
            predictor = train_gpr_model(TRAIN_data)
        else:
            print('No model type:', model_type)

        filename = './params/fold_'+str(i)+'_model.sav'
        print(f'Saving {model_type} model ...')
        pickle.dump(predictor, open(filename, 'wb'))

        #predict on test set
        est=pickle.load(open(filename, 'rb'))
        pred=est.predict(TEST_data.iloc[:,:-1])
        pred = np.array(list(pred))
        gs = np.array(list(TEST_data.iloc[:,-1]))

        pred_all.extend(list(pred))
        gs_all.extend(list(gs))

        auroc = compute_auroc(pred,gs)
        auprc = compute_auprc(pred,gs)
        spearman_cor, _ = spearmanr(pred,gs)
        pearson_cor, _ = pearsonr(pred,gs)
        print("AUROC =", auroc, "AUPRC =", auprc,"Pearson's r =",pearson_cor, "Spearman's r =", spearman_cor)
        # save evaluations
        eva_df['fold'].append(str(i))
        eva_df['AUROC'].append(auroc)
        eva_df['AUPRC'].append(auprc)
        eva_df['Pearsonr'].append(pearson_cor)
        eva_df['Spearmanr'].append(spearman_cor)

    # Overall confidence analysis from k-fold results
    ci = 0.95

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, pearsonr_cor, ci)
    print("Mean[%d%sCI] Pearson's correlation is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['Pearsonr mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, spearmanr_cor, ci)
    print("Mean[%d%sCI] Spearman's correlation is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['Spearmanr mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, compute_auroc, ci)
    print("Mean[%d%sCI] AUROC is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['AUROC mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, compute_auprc, ci)
    print("Mean[%d%sCI] AUPRC is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['AUPRC mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    eva_df = pd.DataFrame.from_dict(eva_df)
    eva_conf_df = pd.DataFrame.from_dict(eva_conf_df)

    return eva_df, eva_conf_df

def transfer_test(df_val):
    """ Transfer validation on validation dataset

    Parameters:
    -----------
    df_val: a Pandas dataframe
        preprocessed validation dataset
        first column: sample
        last column: label

    Yields:
    -------
    eva_df: a Pandas dataframe
        performance of 10 training models on validation dataset
    eva_conf_df: a Pandas dataframe
        confidence evaluations of training models on validation dataset
    """
    path = './validation/'
    os.makedirs(path, exist_ok = True)
    eva_df = {'fold':[],'Pearsonr':[],'Spearmanr':[],'AUROC':[], 'AUPRC':[]}
    eva_conf_df = {'Pearsonr mean[95CI]':[],'Spearmanr mean[95CI]':[],'AUROC mean[95CI]':[], 'AUPRC mean[95CI]':[]}

    TRANSFER = df_val.copy()
    TRANSFER_data = TRANSFER.iloc[:,1:]  # first column is sample name

    for k in range(10):
        filename = './params/fold_'+str(k)+'_model.sav'
        est=pickle.load(open(filename, 'rb'))
        predictions=est.predict(TRANSFER_data.iloc[:,:-1])
        TRANSFER['pred_'+str(k)] = predictions

        pred=predictions
        gs=TRANSFER['label'].to_list()
        # evaluations of predictions
        spearman_cor, _ = spearmanr(pred,gs)
        pearson_cor, _ = pearsonr(pred,gs)
        auroc = compute_auroc(pred,gs)
        auprc = compute_auprc(pred,gs)
        print("Pearson's r",pearson_cor,"Spearman's r = ", spearman_cor, "AUROC =", auroc, "AUPRC = ", auprc)
        eva_df['fold'].append(k)
        eva_df['Pearsonr'].append(pearson_cor)
        eva_df['Spearmanr'].append(spearman_cor)
        eva_df['AUROC'].append(auroc)
        eva_df['AUPRC'].append(auprc)

    # ensemble the predictions from all 10 models
    gs_all = TRANSFER['label'].to_list()
    pred_all = TRANSFER[['pred_'+str(k) for k in range(10)]].mean(axis = 1).to_list()
    TRANSFER['pred_ensemble'] = pred_all
    TRANSFER.to_csv(path+"Test.csv", index = False)

    # save results to table
    RESULTS = TRANSFER[[TRANSFER.columns[0]]+['pred_'+str(k) for k in range(10)]+['pred_ensemble','label']]
    RESULTS.to_csv(path+"results.csv", index = False)

    # Overall confidence analysis from k-fold results
    ci = 0.95

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, pearsonr_cor, ci)
    print("Mean[%d%sCI] Pearson's correlation is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['Pearsonr mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, spearmanr_cor, ci)
    print("Mean[%d%sCI] Spearman's correlation is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['Spearmanr mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, compute_auroc, ci)
    print("Mean[%d%sCI] AUROC is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['AUROC mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    mb, lb, ub = boostrapping_confidence_interval(pred_all, gs_all, compute_auprc, ci)
    print("Mean[%d%sCI] AUPRC is: %.4f[%.4f, %.4f]" % (int(ci*100), '%', mb, lb, ub))
    eva_conf_df['AUPRC mean[95CI]'].append("%.4f[%.4f, %.4f]" %(mb, lb, ub))

    eva_df = pd.DataFrame.from_dict(eva_df)
    eva_conf_df = pd.DataFrame.from_dict(eva_conf_df)
    return eva_df, eva_conf_df



#!usr/bin/env python3
#author: @rayezh
import os, pickle
import shap
from glob import glob
from tqdm import tqdm
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib_venn import venn2, venn2_circles

def SHAP_analysis(regressor, features, feature_names):
    """ Carry out SHAP analysis of trained regressor on target dataset

    Parameters:
    -----------
    regressor:
    data: Numpy array;
        size: n*m
            n: number of instances
            m: number of features
    feature_names: list of strings
        a list of all  feature names

    Yields:
    -------
    shap_df: Pandas dataframe; n row by m column
        SHAP contribution of all features on all instances
    shap_fig: matplotlib plot file
    """
    shap_values = shap.TreeExplainer(regressor).shap_values(np.array(features))
    shap_df = pd.DataFrame(np.array(shap_values), columns = feature_names)

    shap.summary_plot(shap_values, np.array(features), feature_names = feature_names, show=False)
    shap_fig = plt.gcf()
    plt.close()

    return shap_df, shap_fig

def lauch_SHAP():
    """ Lauch SHAP Analysis on both trainnig and validation datasets
    """
    all_model_path = glob('./params/*.sav')
    training_path = './training/'
    validation_path = './validation/'

    # training datasets
    os.makedirs('./training', exist_ok = True)
    os.makedirs('./SHAP/training', exist_ok = True)
    shap_all = []
    for model_path in all_model_path:
        idx = model_path.split('/')[-1].split('_')[1] # index of fold; range from 0 to 9
        regressor = pickle.load(open(model_path, 'rb'))
        fpath = training_path+'fold_'+str(idx)+'/Test.csv'
        Test = pd.read_csv(fpath)
        Test_X = Test.iloc[:, 1:-1]
        shap_df, shap_fig = SHAP_analysis(regressor, Test_X, list(Test_X.columns))
        shap_df.to_csv('./SHAP/training/SHAP_values_'+str(idx)+'_training.csv', index = False)
        shap_fig.savefig('./SHAP/training/SHAP_importance_'+str(idx)+'.pdf', format='pdf', dpi=1200, bbox_inches='tight')
        shap_all.append(shap_df)

    shap_df = pd.concat(shap_all)
    shap_df.to_csv('./SHAP/training/SHAP_values_all_training.csv', index = False)

    # validation datasets
    os.makedirs('./validation', exist_ok = True)
    os.makedirs('./SHAP/validation', exist_ok = True)
    for model_path in all_model_path:
        idx = model_path.split('/')[-1].split('_')[1] # index of fold; range from 0 to 9
        regressor = pickle.load(open(model_path, 'rb'))
        fpath = validation_path+'Test.csv'
        Test = pd.read_csv(fpath)
        Test_X = Test.iloc[:, 1:-12] # This dataframe contains the prediction labels + ensemble prediction score (+11 columns at the end)
        shap_df, shap_fig = SHAP_analysis(regressor, Test_X, list(Test_X.columns))
        shap_fig.savefig('./SHAP/validation/SHAP_importance_'+str(idx)+'.pdf', format='pdf', dpi=1200, bbox_inches='tight')
        shap_df.to_csv('./SHAP/validation/SHAP_values_'+str(idx)+'_validation.csv', index = False)

def sum_SHAP_values_of_genes():
    """ Calculate the SHAP contributions of each gene in predictions (for both training and validation datasets)
        Sort genes by SHAP values
    """

    # training set top genes
    df_training = {'gene':[], 'SHAP_val':[], 'fold':[]}
    paths = glob('./SHAP/training/SHAP_values_*_training.csv')
    for path in paths:
        data = pd.read_csv(path, header = 0)
        idx = path.split('_')[2] # index of fold; range from 0 to 9 and all
        for col in tqdm(data.columns, total = len(data.columns.to_list())):
            gene = col
            shap_val = sum([abs(i) for i in data[col]])/len(data[col])
            df_training['gene'].append(gene)
            df_training['SHAP_val'].append(shap_val)
            df_training['fold'].append(idx)
    df_training = pd.DataFrame.from_dict(df_training)
    df_training = df_training.sort_values(by = 'SHAP_val', ignore_index = True, ascending=False)
    df_training.to_csv('./SHAP/training/SHAP_training_top_genes.csv', index = False)
    df_training_mean = df_training.groupby(['gene'])[['SHAP_val']].mean().reset_index()
    df_training_mean = df_training_mean.sort_values(by = 'SHAP_val', ignore_index = True, ascending=False)
    df_training_mean.to_csv('./SHAP/training/SHAP_training_top_genes_all_fold.csv', index = False)

    # validation set top genes
    df_validation = {'gene':[], 'SHAP_val':[], 'rep':[]}
    fpaths = glob('./SHAP/validation/SHAP_values_*_validation.csv')
    for p in fpaths:
        rep = p.split('_')[2]
        print(rep)
        data = pd.read_csv(p, header = 0)
        for col in tqdm(data.columns, total = len(data.columns.to_list())):
            gene = col
            shap_val = sum([abs(i) for i in data[col]])/len(data[col])
            df_validation['gene'].append(gene)
            df_validation['SHAP_val'].append(shap_val)
            df_validation['rep'].append(rep)
    df_validation = pd.DataFrame.from_dict(df_validation)
    df_validation.to_csv('./SHAP/validation/SHAP_validation_top_genes.csv', index = False)
    df_validation_mean = df_validation.groupby(['gene'])[['SHAP_val']].mean().reset_index()
    df_validation_mean = df_validation_mean.sort_values(by = 'SHAP_val', ignore_index = True, ascending=False)
    df_validation_mean.to_csv('./SHAP/validation/SHAP_validation_top_genes_all_fold.csv', index = False)

def intersect_of_top_genes(n):
    """ Get the intersection of top genes from training and validation sets after performing SHAP analysis.

    Parameters:
    -----------
    n: int
        the top n genes from models to find the intersection of
    """
    # Using the average SHAP value of the gene across 10 models (from 10-fold cross-validation)
    df_training_mean = pd.read_csv('./SHAP/training/SHAP_training_top_genes_all_fold.csv')
    top_training = list(df_training_mean.loc[:(n-1), 'gene'])

    df_validation_mean = pd.read_csv('./SHAP/validation/SHAP_validation_top_genes_all_fold.csv')
    top_val = list(df_validation_mean.loc[:(n-1), 'gene'])

    # Find genes in common among the top n genes from each set
    common = list(set(top_training) & set(top_val))
    # Save that to a file for future reference
    with open('./SHAP/intersection_list_top_' + str(n) + '_genes.txt', 'w') as file:
        for gene in common:
            file.write("%s\n" % gene)

    # Make Venn diagram
    n_training = n - len(common)
    n_val = n - len(common)

    plt.figure(figsize=(5,4))
    v = venn2(subsets = (n_training, n_val, len(common)), set_labels = ('Training', 'Validation'), set_colors=('yellowgreen','orange'), alpha = 0.7)
    c = venn2_circles(subsets = (n_training, n_val, len(common)), linewidth=1, color='k')
    plt.title('Intersection of top %d genes from SHAP analysis\n on training and validation sets' % n)
    plt.savefig('./SHAP/intersection_venn_top_' + str(n) + '_genes.pdf', format='pdf', dpi=1200, bbox_inches='tight')
    plt.close()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ciclops",
    version="0.0.4",
    author="Elysia Chou, Hanrui Zhang, Yuanfang Guan",
    author_email="elysian@umich.edu, rayezh@umich.edu, gyuanfan@umich.edu",
    description="Pipeline for building clinical outcome prediction models on training dataset and transfer learning on validation datasets.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GuanLab/ciclops",
    project_urls={
        "Bug Tracker": "https://github.com/GuanLab/ciclops/issues",
    },
    keywords=['Bioinformatics','Transfer Learning','Machine Learning','Transcriptomics','Research', 'Clinical Outcome Prediction'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(where = 'src'),
    entry_points = {
        'console_scripts': [
            'ciclops = ciclops.__main__:main'
            ]
        },
    python_requires=">=3.8",
    install_requires = [
        'numpy >=1.21.5',
        'pandas >=1.4.1',
        'matplotlib >=3.5.1',
        'matplotlib-venn >=0.11.7',
        'scikit-learn >=1.0.2',
        'scipy >=1.8.0',
        'lightgbm >=3.3.2',
        'shap >=0.40.0',
        'xgboost >=1.6.0',
        'tqdm >=4.63.0'
    ]
)



from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import precision_recall_curve, roc_curve, auc
import pandas as pd
import numpy as np
def pearsonr_cor(pred, label):
    """ Return Pearson's correlation between prediction and label
    """
    cor, _ = pearsonr(pred, label)
    return cor

def spearmanr_cor(pred, label):
    """ Return Spearman's correlation between prediction and label
    """
    cor, _ = spearmanr(pred, label)
    return cor

def compute_auroc(pred, label):
    """ Calculate AUROC of predictions in terms of label
    """
    #label = np.array(label)
    #pred = np.array(pred)
    fpr, tpr, thresholds = roc_curve(label, pred, pos_label =1)
    auroc = auc(fpr, tpr)
    return auroc

def compute_auprc(pred, label):
    """ Calculate AUPRC of predictions in terms of label
    """
    #label = np.array(label)
    #pred = np.array(pred)
    precision, recall, thresholds = precision_recall_curve(label, pred)
    auprc = auc(recall, precision)
    return auprc


def c_index(pred, label):
    """ Compute C-idex of the predictions in terms of ground truth
    
    Parameters:
    -----------
    pred: list
        prediction
    label: list
        ground truth
    pred and label are the same length
    
    Yields:
    -------
    cidx: float
        C-index (between 0 to 1)
    """
    from itertools import permutations
    pred = list(pred)
    label = list(label)
    perm = permutations(list(range(len(pred))), 2)
    survive = 0
    total = 0
    for i, j in perm:
        if label[i]<label[j]:
            total +=1
            if pred[i]<pred[j]:
                survive += 1
    cidx = survive/total
    return cidx

def boostrapping_confidence_interval(pred_all, gs_all, eva_func, ci):
    """ Boostrapping to get a 95 confidence interval for prediction performance
    Parameters:
    -----------
    pred_all: list
        all predictions from k-fold cross-validations
    gs_all: list
        all gold standards from k-fold cross-validations
    eva_func: function
        evaludation function
    ci: confidence interval

    Yields:
    -------
    mb: float
        middle bound
    lb: float
        lower bound
    ub: float
        upper bound
    """
    import numpy as np
    import random
    # set random seed
    random.seed(0)

    # prediction-groundtruth pairs from all five fold cross validation
    tmp = np.array([pred_all, gs_all]).T
    # calculate overall correlation
    mb = eva_func(tmp[:,0], tmp[:,1])
    # start boostrapping ...
    eva_all = []
    for i in range(100):
        tmp_new = random.choices(tmp, k = len(tmp))
        tmp_new = np.array(tmp_new)
        eva = eva_func(tmp_new[:,0], tmp_new[:,1])
        eva_all.append(eva)
    eva_all = sorted(eva_all)
    #print(eva_all)
    lb = eva_all[round(100*(0.5-ci*0.5))]
    ub = eva_all[round(100*(0.5+ci*0.5))]
    return mb, lb, ub
